/**
 * This class is used to create a Client object.
 * The client handels the client server communication.
 * It requests data (e.g. the analysis session or query results) from the
 * the server and recieves them, if it dynamicMode is true.
 * If dynamicMode is false, the client can also handle data that is already
 * loaded and not coming in in chunks.
 */
class Client {
  constructor({corpusId = null,
               socket = null,
               logging = true,
               dynamicMode = true} = {}) {
    this.corpusId = corpusId;
    this.displays = {};
    this.dynamicMode = dynamicMode;
    this.logging = logging;
    this.requestQueryProgress = 0;
    this.results = undefined;  // holds the results object later on
    this.socket = socket;
    this.socketEventListeners = {};

    /**
     * Disable all console logging.
     * Credits to https://gist.github.com/kmonsoor/0244fdb4ad79a4826371e58a1a5fa984
     */
    if (!logging) {
      (() => {
        let console = (window.console = window.console || {});
        [
          'assert', 'clear', 'count', 'debug', 'dir', 'dirxml',
          'error', 'exception', 'group', 'groupCollapsed', 'groupEnd',
          'info', 'log', 'markTimeline', 'profile', 'profileEnd', 'table',
          'time', 'timeEnd', 'timeStamp', 'trace', 'warn'
        ].forEach((method) => {
          console[method] = () => {};
        });
      })();
    }
  }

  // Registers one or more SocketEventListeners to the Client.
  setSocketEventListeners(socketEventListeners) {
    for (let socketEventListener of socketEventListeners) {
      this.socketEventListeners[socketEventListener.type] = socketEventListener;
    }
  }

  /**
   * Loads the SocketEventListeners so they will be triggered on their assigned
   * type strings because they double as the socket event event names.
   */
  loadSocketEventListeners() {
    for (let [type, listener] of Object.entries(this.socketEventListeners)) {
      listener.listenerFunction(type, this);
    }
  }


  // TODO: get rid of this disply stuff and send commands to the viewer
  // on what to do/show/hide etc

  notifyView(SendWhatToDo) {

  }

  // Registers a CorpusAnalysisDisplay object to the Client.
  setDisplay(type, corpusAnalysisDisplay) {
    this.displays[type] = corpusAnalysisDisplay;
  }

  /**
   * Function that takes one or more query selector
   * strings in an array as an input. The function then creates a
   * class field in the client object with the query selector
   * string as the key. The selector will be converted to a valid JavaScript
   * Field name i. e. #html-id-string -> this.htmlIdString
   * The value will be the identifed element fetched with the querySelector
   * method.
   */
  // TODO: multipleResults=false, atattchSomeCallback=false ?
  getHTMLElements(arrayOfSelectors) {
    for (let selector of arrayOfSelectors) {
      let element = document.querySelector(selector);
      let cleanKey = [];
      selector = selector.replace('_', '-');
      selector.match(/\w+/g).forEach((word) => {
        let tmp = word[0].toUpperCase() + word.slice(1);
        cleanKey.push(tmp);
      });
      cleanKey[0] = cleanKey[0].toLowerCase();
      cleanKey = cleanKey.join('');
      this[cleanKey] = element;
    }
  }

  /**
   * Connects to the corpus analysis session for the specified corpus via
   * socket.io.
   */
  connect() {
    console.info('corpus_analysis_init: Client connecting to session via',
                 'socket.emit');
    // if (this.displays.init != undefined) {
    //   this.displays.init.element.M_Modal.open();
    //   this.displays.init.setVisibilityByStatus('waiting');
    // }
    this.socket.emit('corpus_analysis_init', this.corpusId);
  }

  /**
   * Emits query to the server via socket.io. Server will send the results
   * back.
   */
  query(queryStr) {
    console.info('corpus_analysis_query: Client sending query via',
                 'socket.emit for the query', queryStr);
    this.socket.emit('corpus_analysis_query', queryStr);
  }
}

/**
 * This class is used to create an CorpusAnalysisDisplay object.
 * Input is one HTMLElement that can then be hidden or shown depending on
 * its CSS classes.
 */
class CorpusAnalysisDisplay {
  constructor(element) {
    // with this function initalized modals can also be handeld
    this.element = (() => {if (element instanceof HTMLElement) {
                            return element;
                          } else {
                            element = element['$el'][0];
                            return element;
                          }
    })();
    this.errorContainer = element.querySelector('.error-container');
    this.showOnError = element.querySelectorAll('.show-on-error');
    this.showOnSuccess = element.querySelectorAll('.show-on-success');
    this.showWhileWaiting = element.querySelectorAll('.show-while-waiting');
    this.hideOnComplete = element.querySelectorAll('.hide-on-complete')
  }

  // Changes the visibility of its own
  setVisibilityByStatus(status) {
    switch (status) {
      case 'error':
        for (let element of this.showOnError) {
          element.classList.remove('hide');
        }
        for (let element of this.showOnSuccess) {
          element.classList.add('hide');
        }
        for (let element of this.showWhileWaiting) {
          element.classList.add('hide');
        }
        break;
      case 'success':
        for (let element of this.showOnError) {
          element.classList.add('hide');

        }
        for (let element of this.showOnSuccess) {
          element.classList.remove('hide');
        }
        for (let element of this.showWhileWaiting) {
          element.classList.add('hide');
        }
        break;
      case 'waiting':
        for (let element of this.showOnError) {
          element.classList.add('hide');
        }
        for (let element of this.showOnSuccess) {
          element.classList.add('hide');
        }
        for (let element of this.showWhileWaiting) {
          element.classList.remove('hide');
        }
        break;
      default:
        // Hide all
        for (let element of this.showOnError) {
          element.classList.add('hide');
        }
        for (let element of this.showOnSuccess) {
          element.classList.add('hide');
        }
        for (let element of this.showWhileWaiting) {
          element.classList.add('hide');
        }
    }
  }
}

/**
 * This class is used to create an SocketEventListener.
 * Input are an identifying type string, the listener function and callbacks
 * which will be executed as part of the listener function. The identifying
 * type string is also used as the socket event event identifier.
 */
class SocketEventListener {
  constructor(type, listenerFunction) {
    this.listenerCallbacks = {};
    this.listenerFunction = listenerFunction;
    this.type = type;
  }

  // Registers callbacks to this SocketEventListener
  setCallbacks(listenerCallbacks) {
    for (let listenerCallback of listenerCallbacks) {
      this.listenerCallbacks[listenerCallback.type] = listenerCallback;
    }
  }

  /** Shorthand to execute all registered callbacks with same args in insertion
   * order.
   * NOTE:
   * Since ECMAScript 2015, objects do preserve creation order for
   * string and Symbol keys. In JavaScript engines that comply with the
   * ECMAScript 2015 spec, iterating over an object with only string keys will
   * yield the keys in order of insertion.
   * So all modern Browsers.
   */
  executeCallbacks(args) {
    for (let [type, listenerCallback] of Object.entries(this.listenerCallbacks)) {
      listenerCallback.callbackFunction(...args);
    }
  }
}

/**
 * This class is used to create an ListenerCallback which will be registered
 * to an SocketEventListener so the Listener cann invoke the ListenerCallback
 * callback functions.
 */
class ListenerCallback {
  constructor(type, callbackFunction) {
    this.callbackFunction = callbackFunction;
    this.type = type;
  }
}

// export Classes from this module
export {
  Client,
  SocketEventListener,
  CorpusAnalysisDisplay,
  ListenerCallback,
};